#ifndef TANGO_IR_HPP
#define TANGO_IR_HPP

#include "tango_globals.hpp"
#include "tango_inst.hpp"
#include "code_buffer.hpp"
#include "stacks.hpp"
#include "tango_globals.hpp"
#include "utilities/bitMaps.hpp"
#include "tango_allocator.hpp"
#include "cpuinfo.hpp"
#include <string>

const float SMALL_PROBABILITY = 1E-6f;
const float BIG_PROBABILITY = 1.0f - SMALL_PROBABILITY;

const float BC_CLONE_LIMIT = 1.0f;
const float BC_CLONE_PROB_THRESHOLD = 0.6f;

const int SCH_ST_LOCK_CROSSES_LIMIT = 10; // Max number of locks which can intersect one store
const int SCH_LC_INTERSECTION_LIMIT = 2; // max number of intersecting stores which lock can cross
const int SCH_RBRANCH_COMP_CODE_LIMIT = 12; // max inst cnt which can be moved through check insts

class OptoScopeInfo;

namespace tango {
const int TangoLoopUnrollLimit = 90;  // loop body size limit for unroll

class TangoScopeInfo;

//----------------------------------------------------------------------------//

union DoubleInteger {
  uint64_t i;
  double   d;
};

//----------------------------------------------------------------------------//

enum FailCond {
  NoFail,
  RegAllocFail,
  SelectorFail
};

enum class PeepHoleMode: uint8_t {
  FMA,
  CMPIBRANCH
};

// Enum to denote special positions in use lists of instructions
namespace OpndPos {
enum {
  // All instructions have a predicate as the first use
  OpndPred = 0,
  // Load uses, load use list looks like this:
  // ldx pred, base, offset
  OpndLddAddr = 1,
  OpndLddOffset = 2,

  // Store uses, store use list looks like this:
  // stx pred, value, base, offset
  OpndStdValue = 1,
  OpndStdAddr = 2,
  OpndStdOffset = 3,

  // Branch targets
  OpndBranchTrg = 1,
  OpndCtCondGotoTrg = 1,
  OpndCtCondCondTrg = 2,

  // Mov value
  OpndMovVal = 1,

  // Merge uses, the merge uses look like this:
  // 0. Pred - predicate for the whole instruction
  // 1. Merge pred - predicate that controls the merge of two values
  // 2. True val - value that is selected if merge pred is true
  // 3. False val - value that is selected if merge pred is false
  OpndMergePred = 1,
  OpndMergeTrue = 2,
  OpndMergeFalse = 3
};
}

// Relocations

class RelocLts;
class RelocSafepoint;
class RelocCodeOffset;
class RelocLDisp;
class RelocDisp;
class DebugInfoTransfer;

class Ir;

class Reloc : public TangoObj {
public:
  enum RelocInfo {
    NONE,
    LTS,
    SAFEPOINT,
    CODE_OFFSET,
    LDISP,
    DISP
  };

protected:
  void*     _address;
  int       _index;  //For links between relocations

private:
  RelocInfo _type;

protected:
  Reloc(RelocInfo _type, Ir* ir);

public:
  virtual Reloc* clone(Allocator *alloc, Ir* ir = NULL, Reloc* link = NULL);

  void      setAddress(void *address) { _address = address; }
  RelocInfo type() const              { return _type; }
  void*     addr() const              { return _address; }
  int       index() const             { return _index; }

  virtual int getFormat() const       { return 0; }
  virtual RelocSafepoint* asRelocSafepoint()  { return nullptr; }
  virtual RelocLts* asRelocLts()  { return nullptr; }
  virtual RelocCodeOffset* asRelocCodeOffset() { return nullptr; }
  virtual RelocLDisp* asRelocLDisp() { return nullptr; }
  virtual RelocDisp* asRelocDisp() { return nullptr; }
};

// Base class for imm relocations such as Oop, NarrowOpp, Metadata, CodePc, etc...

class RelocLts : public Reloc {
public:
  RelocLts(Ir* ir) : Reloc(LTS, ir) { }
  virtual uint64_t onRelocLts(uint64_t value, DebugInfoTransfer *debugInfoTransfer) { return value; /* do nothing by default */ }
  RelocLts* asRelocLts()  { return this; }
};

// Base class for special relocations such as E2kCtReloc

class RelocSafepoint : public Reloc {
public:
  RelocSafepoint(Ir* ir) : Reloc(SAFEPOINT, ir) { }
  virtual void onSafepoint(uint32_t disp, uint32_t disp_next, Ir* ir) { }
  RelocSafepoint* asRelocSafepoint()  { return this; }
};

// Code offset relocations

class RelocCodeOffset : public Reloc {
public:
  RelocCodeOffset(Ir* ir) : Reloc(CODE_OFFSET, ir) { }
  RelocCodeOffset* asRelocCodeOffset() { return this; }
};

// ldisp stores displacement between this ldisp bundle and fapb (written to const section)
// this relocation is used to update displacement after code compacting
class RelocLDisp: public Reloc {
public:
  uint32_t c0_offset;      // offset from insts start to c0 syllable which will be patched
  uint32_t async_offset;   // offset of fapb operations from async first fapb
  uint32_t insts_offset;   // offset of ldisp operation from insts start
public:
  RelocLDisp(Ir* ir, uint32_t _c0_off, uint32_t _async_offset, uint32_t _disp_offset):
    Reloc(LDISP, ir),
    c0_offset(_c0_off), async_offset(_async_offset), insts_offset(_disp_offset) {}

  Reloc*           clone(Allocator *alloc, Ir* ir = NULL, Reloc* link = NULL);
  void             set_c0_offset(uint32_t c0_off) { c0_offset = c0_off; }
  uint32_t         get_c0_offset() const          { return c0_offset;   }

  RelocLDisp*      asRelocLDisp()                 { return this; }
};

class RelocDisp : public Reloc {
public:
  uint32_t c0_offset; // offset from bundle start to c0 syllable which will be patched
public:
  RelocDisp(Ir* ir) : Reloc(DISP, ir) { }
  virtual void onRelocDisp(Ir* ir) { /* do nothing by default */ }
  RelocDisp* asRelocDisp()  { return this; }
};

//----------------------------------------------------------------------------//

class ScopeInfo : public TangoObj {
public:
  virtual bool isTangoScopeInfo() const { return false; }
  virtual bool isOptoScopeInfo() const { return false; }
  TangoScopeInfo* asTangoScopeInfo() const { assert_rvm(isTangoScopeInfo(), "verufy type"); return (TangoScopeInfo*)this; }
  OptoScopeInfo*  asOptoScopeInfo()  const { assert_rvm(isOptoScopeInfo(),  "verufy type"); return (OptoScopeInfo*)this; }
public:
  //OopMap* oopmap;
  Inst* inst;
  int pc_offset;
};

struct DerivedPair {
  DerivedPair() : ptr(NULL), base(NULL) {}
  DerivedPair(Opnd* p, Opnd* b) : ptr(p), base(b) {
    assert_rvm(base != NULL, "derived must have base");
  }

  Opnd* ptr;
  Opnd* base;
};

struct ExceptionTableEntry : public TangoObj {
  arena_vector<intptr_t>::type bcis;
  BlockVector blocks;
  Inst* catch_inst;
};

inline bool operator<(const DerivedPair& l, const DerivedPair& r) { return l.ptr < r.ptr; }

typedef arena_map<int, int>::type       CallSiteMap;
typedef arena_vector<ScopeInfo*>::type  ScopeVector;
typedef arena_set<DerivedPair>::type    DerivedRegsSet;
typedef arena_map<Inst*, ExceptionTableEntry*>::type ExceptionTableMap;

class SafepointInfo {
public:
  CallSiteMap& call_site_map() { return call_site_map_; }
  ScopeVector& scopes() { return scope_vector_; }
  DerivedRegsSet& derived_regs() { return derived_regs_; }

private:
  CallSiteMap call_site_map_;
  ScopeVector scope_vector_;
  DerivedRegsSet derived_regs_;
};

//----------------------------------------------------------------------------//
// Class is used to carry debug info and relocations through various phases of compilation

class DebugInfoTransfer : public TangoObj {

  struct RegistrationPoint {
    uint32_t offset;
    uint64_t address;
  };

public:
  DebugInfoTransfer(Ir* ir_) : ir(ir_) { }
  uint32_t  getCallByOffset(uint32_t offset);
  uint64_t  getAddrByCall(uint32_t call_inst_id);
  uint32_t  getOffsetByCall(uint32_t call_inst_id);
  uint64_t  getRealOffset(uint32_t selector_offset);
  void      addAddrToPatch(uint32_t call, uint64_t addr, uint32_t offset);
  void      addRealOffset(uint32_t selector_offset, uint64_t offset);
  void      registerCall(uint32_t offset, uint32_t call_inst_id);
  void      addNewReloc(void* addr, Reloc *reloc);

  DerivedRegsSet& derivedRegs()    { return safepointInfo.derived_regs(); }
  CallSiteMap&    callSiteMap()    { return safepointInfo.call_site_map(); }
  ScopeVector&    scopes()         { return safepointInfo.scopes(); }

  RelocVector relocs;
  SafepointInfo safepointInfo;
  ExceptionTableMap xTable;

private:
  arena_map<uint32_t, uint32_t>::type offset_to_call;
  arena_map<uint32_t, RegistrationPoint>::type call_to_offset;
  arena_map<uint32_t, uint64_t>::type offset_to_offset;
  Ir *ir;
};

class Pragmas : public TangoObj {
public:
  Pragmas() : _pragmas(0), _unroll_value(-1) { }

  void addIVDep()   { _pragmas |= LOOP_IVDEP  ; }
  void addUnroll()  { _pragmas |= LOOP_UNROLL ; }
  void addUncheck() { _pragmas |= LOOP_UNCHECK; }
  void addFFast()   { _pragmas |= LOOP_FFAST  ; }
  void addSimd()    { _pragmas |= LOOP_SIMD   ; }

  bool hasIVDep()   const { return _pragmas & LOOP_IVDEP   != 0; }
  bool hasUnroll()  const { return _pragmas & LOOP_UNROLL  != 0; }
  bool hasUncheck() const { return _pragmas & LOOP_UNCHECK != 0; }
  bool hasFFast()   const { return _pragmas & LOOP_FFAST   != 0; }
  bool hasSimd()    const { return _pragmas & LOOP_SIMD    != 0; }

  void setUnrollValue(int v) { assert_rvm(hasUnroll(), "sanity"); _unroll_value = v; }
  int  getUnrollValue()      const { assert_rvm(hasUnroll(), "sanity"); return _unroll_value; }

protected:
  int   _pragmas;
  int   _unroll_value;
};

//----------------------------------------------------------------------------//

class ImmOpnd;
class RegOpnd;
class LabelOpnd;
class SymbolOpnd;

class Opnd : public TangoObj {
public:
  Opnd(int32_t, DataType, uint32_t);
  int32_t          getSize()      const { return Globals::getDataSize(data_type); }
  bool             isSigned()     const { return Globals::isSigned(data_type); }
  virtual bool     isImm()        const { return false; }
  virtual bool     isReg()        const { return false; }
  virtual bool     isRreg()       const { return false; }
  virtual bool     isBRreg()      const { return false; }
  virtual bool     isGreg()       const { return false; }
  virtual bool     isPreg()       const { return false; }
  virtual bool     isBPreg()      const { return false; }
  virtual bool     isCreg()       const { return false; }
  virtual bool     isAAreg()      const { return false; }
  virtual bool     isSreg()       const { return false; }
  virtual bool     isLabel()      const { return false; }
  virtual bool     isSymbol()     const { return false; }
  ImmOpnd*    asImm()        const { assert_rvm(isImm(), "check"); return (ImmOpnd*)this; }
  RegOpnd*    asReg()        const { assert_rvm(isReg(), "check"); return (RegOpnd*)this; }
  LabelOpnd*  asLabel()      const { assert_rvm(isLabel(), "check"); return (LabelOpnd*)this; }
  SymbolOpnd* asSymbol()     const { assert_rvm(isSymbol(), "check"); return (SymbolOpnd*)this; }
  virtual uint64_t getValue()     const { TANGO_UNREACHABLE; return 0; }  // opnd value for encoder
  virtual uint64_t getValueOld()  const { TANGO_UNREACHABLE; return 0; }
  bool             isExactSize()  const { return (flags & OF_EXACT_SIZE)  != 0; }
  bool             isManagedPtr() const { return (flags & OF_MANAGED_PTR) != 0; }
  bool             isNarrowPtr()  const { return (flags & OF_NARROW_PTR)  != 0; }
  bool             isDerivedPtr() const { return (flags & OF_DERIVED_PTR) != 0; }
  bool             isOutArg()     const { return (flags & OF_OUT_ARG)     != 0; }
  bool             isInArg()      const { return (flags & OF_IN_ARG)      != 0; }
  bool             isSpillable()  const { return (flags & OF_UNSPILLABLE) == 0; }
  bool             isValue()      const { return (flags & OF_VALUE)       != 0; }
  bool             isInverted()   const { assert_rvm(isPreg(), "avaliable only for preds"); return (flags & OF_INVERTION)   != 0; }
  bool             isInt()        const;
  bool             isFp()         const;
  bool             is_jobject()   const; // TODO
  bool             isNonZero()    const { return (flags & OF_NON_ZERO)    != 0; }
  void             markManagedPtr() { flags |= OF_MANAGED_PTR; }
  void             setNonZero(bool b) { if (b) { flags |= OF_NON_ZERO; } else { flags &= ~OF_NON_ZERO; } }
  void             setExactSize(bool b) { if (b) { flags |= OF_EXACT_SIZE; } else { flags &= ~OF_EXACT_SIZE; } }
  void             setInvertion(bool b) { assert_rvm(isPreg(), "avaliable only for preds"); if (b) { flags |= OF_INVERTION;  } else { flags &= ~OF_INVERTION;  } }
  void             setUnspillable(bool b) { assert_rvm(isReg(), "available only for vregs"); if (b) { flags |= OF_UNSPILLABLE;  } else { flags &= ~OF_UNSPILLABLE;  } }

public:
  int32_t   id;
  DataType  data_type;
  uint32_t  flags;              // OF_OUT_ARG OF_IN_ARG OF_MANAGED_PTR
  int32_t   local_idx;
  Reloc*    reloc;
};


//----------------------------------------------------------------------------//

class ImmOpnd : public Opnd {
public:
  ImmOpnd(int32_t id, DataType dt, uint64_t i, uint32_t flags = 0) : Opnd(id, dt, flags) { imm.i = i; }
  ImmOpnd(int32_t id, DataType dt, double   d, uint32_t flags = 0) : Opnd(id, dt, flags) { imm.d = d; }

  bool      isImm()             const { return true; }
  uint64_t  getValue()          const { return imm.i; }
  uint64_t  getValueOld()       const { return imm.i; }
  uint64_t  getInt()            const { return imm.i; }
  double    getDouble()         const { return imm.d; }
  void      setInt(uint64_t i)        { imm.i = i; }
  void      setDouble(double d)       { imm.d = d; }
  bool      fitsInSize(int32_t) const;
  int32_t   fitsInSize()        const;
  bool      isPackable() { return imm.i & 0xffffffffffffffe0 == 0 ? true : false; }

public:
  DoubleInteger imm;
};

//----------------------------------------------------------------------------//

class LabelOpnd : public Opnd {
public:
  LabelOpnd(int32_t id, Block* block, float prob = 1.0f) : Opnd(id, DATA_U64, 0), block(block), prob(prob) {}
  bool isLabel()         const { return true; }
  uint64_t getValue()    const { return 0; }
  uint64_t getValueOld() const { return 0; }

public:
  Block* block;
  float prob;
};

//----------------------------------------------------------------------------//

class SymbolOpnd : public Opnd {
public:
  SymbolOpnd(int32_t id, std::string str) : Opnd(id, DATA_U64, 0), str(str) {}
  bool isSymbol() const { return true; }
  uint64_t getValue()     const { return 0; }

public:
  std::string str;
};

//----------------------------------------------------------------------------//

class RegOpnd : public Opnd {
public:
  RegOpnd(int32_t, RegType, DataType, int32_t, uint32_t=0, bool=true);
  bool     isReg()       const { return true; }
  bool     isRegNum()    const { return location >= EREG_G0 && location < EREG_LAST; }
  bool     isStackSlot() const { return location >= EREG_LAST; }
  bool     isAllocated() const { return location != UNALLOCATED; }
  bool     isRreg()      const { return reg_type == REG_TYPE_R; }
  bool     isBRreg()     const { return location >= EREG_BR0 && location <= EREG_BR127; }
  bool     isPreg()      const { return reg_type == REG_TYPE_PR; }
  bool     isBPreg()     const { return location >= EREG_BPR0 && location <= EREG_BPR31; }
  bool     isCreg()      const { return reg_type == REG_TYPE_CTPR; }
  bool     isAAreg()     const { return reg_type == REG_TYPE_AAD || reg_type == REG_TYPE_AASTI || reg_type == REG_TYPE_AAIND
        || reg_type == REG_TYPE_AAINCR || reg_type == REG_TYPE_AALDI || reg_type == REG_TYPE_AALDI || reg_type == REG_TYPE_AASPEC; }
  bool     isSreg()      const { return reg_type == REG_TYPE_SPECIAL; }
  bool     isEmpty()     const { return reg_type == REG_TYPE_SPECIAL && location == EREG_EMPTY_HI;}
  bool     isGreg()      const { return location >= EREG_G0 && location <= EREG_G31; }
  bool     isLocSlot()   const;
  RegOpnd* addUse(Inst*);
  RegOpnd* addDef(Inst*);
  RegOpnd* delUse(Inst*);
  RegOpnd* delDef(Inst*);
  uint64_t getValue()    const;
  uint64_t getValueOld() const;

public:
  RegType  reg_type;
  int32_t  location;           // 0 to EREG_LAST - reg nums, then - stack slots

  bool     cross_call_site;    // opnd live range crosses call inst
  double   spill_cost;         // cost of spilling this opnd on stack
  int32_t  dep_id;             // index of this opnd in dep_set
  BitMap   dep_set;            // opnds alive with this one
  BitMap   free_reg_mask;      // regs available for this opnd (filled by registed allocator)
  int32_t  als_id;             // last ALU number, that contains def of this opnd
  Inst*    last_def;
};

//----------------------------------------------------------------------------//

// VM-specific structure, contains some high-level IR information for every tango inst
// Used for memory disambiguations
class MemoryInfo : public TangoObj {
public:
  virtual bool has_mismatched_access() const { return false; }
  virtual bool canMemoryIntersect(MemoryInfo*) const = 0;
  virtual bool exactlySame(MemoryInfo*) const = 0;
};

//----------------------------------------------------------------------------//

class Inst : public TangoObj {
public:
  Inst(int32_t id, InstCode opcode) : id(id), opcode(opcode), flags(0U), use_clp_num(0U), linked_opcode(INST_LAST), mem_info(NULL), als_id(ILLEGAL_ALS_ID), entry((address)0xDEAD), size(-1) {
      uses.reserve(4); // Default capacity is 0. To avoid re-allocation do reserve beforehand.
      touch_uses.clear();
  }

  Inst*           clone(Allocator *a) const;
  Inst*           copy(Allocator *a) const;
  Inst*           full_copy(Allocator *a) const;
  Inst*           addUse(Opnd* opnd);
  Inst*           addTouchUse(Opnd* opnd);
  Inst*           addDef(Opnd* opnd);
  int8_t          getLatency(Inst* trg, bool bypass = true, int CPUModel = 0)  const {
    return Globals::getLatency(opcode, trg->opcode, bypass, CPUModel);
  }
  uint32_t          getEuMask()            const { return Globals::getEuMask(opcode); }
  const char*       getName()              const { return Globals::inst_info[opcode].name; }
  bool              isMemRead()            const { return Globals::isMemRead(opcode); }
  bool              isMemWrite()           const { return Globals::isMemWrite(opcode); }
  bool              isAAInst()             const { return opcode >= INST_LDAAB && opcode <= INST_AAURWD; }
  bool              isMova()               const { return opcode >= INST_MOVAB && opcode <= INST_MOVAD; }
  bool              isStaa()               const { return opcode >= INST_STAAB && opcode <= INST_STAAD; }
  bool              isStWait()             const;
  bool              isLdWait()             const;
  bool              isClf()                const;
  bool              isNullCheck()          const;
  bool              hasIbranch()           const { return opcode == INST_IBRANCH || opcode == INST_MCMPIBRANCH || opcode == INST_MOVIBRANCH; }
  bool              isProcedureClf()       const { return (opcode == INST_CALL || opcode == INST_CT_RET); }
  bool              isPacked()             const { return opcode >= INST_PADDW && opcode <= INST_PSHUFH; }
  Inst*             setInverse(bool b)           { if (b) { flags |= IF_PRED_INVERSE; } else { flags &= ~IF_PRED_INVERSE; } return this; }
  Inst*             setSpeculative(bool b)       { if (b) { flags |= IF_SPECULATIVE; } else { flags &= ~IF_SPECULATIVE; } return this; }
  Inst*             setPossibleException(bool b) { if (b) { flags |= IF_POSSIBLE_EXC; } else { flags &= ~IF_POSSIBLE_EXC; } return this; }
  Inst*             setSafepoint(bool b)         { if (b) { flags |= IF_SAFEPOINT; } else { flags &= ~IF_SAFEPOINT; } return this; }
  Inst*             setUse1Inverted(bool b)      { if (b) { flags |= IF_USE1_INVERT; } else { flags &= ~IF_USE1_INVERT; } return this; }
  Inst*             setUse2Inverted(bool b)      { if (b) { flags |= IF_USE2_INVERT; } else { flags &= ~IF_USE2_INVERT; } return this; }
  Inst*             setUse3Inverted(bool b)      { if (b) { flags |= IF_USE3_INVERT; } else { flags &= ~IF_USE3_INVERT; } return this; }
  Inst*             setUse4Inverted(bool b)      { if (b) { flags |= IF_USE4_INVERT; } else { flags &= ~IF_USE4_INVERT; } return this; }
  Inst*             setUse5Inverted(bool b)      { if (b) { flags |= IF_USE5_INVERT; } else { flags &= ~IF_USE5_INVERT; } return this; }
  Inst*             setUse6Inverted(bool b)      { if (b) { flags |= IF_USE6_INVERT; } else { flags &= ~IF_USE6_INVERT; } return this; }
  Inst*             setUncheckedPtrLoad(bool b)  { if (b) { flags |= IF_UNCHECKPTR_LD; } else { flags &= ~IF_UNCHECKPTR_LD; } return this; }
  Inst*             setInvariantLoad(bool b)     { if (b) { flags |= IF_INVARIANT_LD; } else { flags &= ~IF_INVARIANT_LD; } return this; }
  Inst*             setNoSECall(bool b)          { if (b) { flags |= IF_NO_SIDE_EFFECT; } else { flags &= ~IF_NO_SIDE_EFFECT; } return this; }
  Inst*             setSpill(bool b)             { if (b) { flags |= IF_SPILL; } else { flags &= ~IF_SPILL; } return this; }
  Inst*             setSingle(bool b)            { if (b) { flags |= IF_SINGLE; } else { flags &= ~IF_SINGLE; } return this; }
  Inst*             setSideEffects(bool b)       { if (b) { flags |= IF_SIDE_EFFECT; } else { flags &= ~IF_SIDE_EFFECT; } return this; }
  Inst*             setLoadLock();
  Inst*             setLoadCheck();
  Inst*             setLoadDefault();
  bool              isInverse()            const { return ((flags & IF_PRED_INVERSE) != 0); }
  bool              isSpeculative()        const { return ((flags & IF_SPECULATIVE) != 0); }
  bool              isPossibleException()  const { return ((flags & IF_POSSIBLE_EXC) != 0); }
  bool              isSafepoint()          const { return ((flags & IF_SAFEPOINT) != 0); }
  bool              isUncheckedPtrLoad()   const { return ((flags & IF_UNCHECKPTR_LD) != 0); }
  bool              isInvariantLoad()      const { return ((flags & IF_INVARIANT_LD) != 0); }
  bool              isNoSECall()           const { return ((flags & IF_NO_SIDE_EFFECT) != 0); }
  bool              isUse1Inverted()       const { return ((flags & IF_USE1_INVERT) != 0); }
  bool              isUse2Inverted()       const { return ((flags & IF_USE2_INVERT) != 0); }
  bool              isUse3Inverted()       const { return ((flags & IF_USE3_INVERT) != 0); }
  bool              isUse4Inverted()       const { return ((flags & IF_USE4_INVERT) != 0); }
  bool              isUse5Inverted()       const { return ((flags & IF_USE5_INVERT) != 0); }
  bool              isUse6Inverted()       const { return ((flags & IF_USE6_INVERT) != 0); }
  bool              isSpill()              const { return ((flags & IF_SPILL) != 0); }
  bool              isSingle()             const { return ((flags & IF_SINGLE) != 0); }
  bool              hasSideEffects()       const { return ((flags & IF_SIDE_EFFECT) != 0); }
  bool              isLoadLock()           const { return isMemRead() && mas() == LD_MAS_LOCK_CHECK && isSpeculative(); }
  bool              isLoadCheck()          const { return isMemRead() && mas() == LD_MAS_CHECK_UNLOCK && !isSpeculative(); }
  bool              isRbranch()            const;
  bool              isPredicateDef()       const;
  bool              hasOpndDef(Opnd* op)   const;
  bool              isIntCmp()             const;
  const OpndVector& defs()                 const { return _defs; }
  OpndVector&       unsafeDefs()                 { return _defs; }
  Opnd*             getUse(int32_t)        const;
  Opnd*             getDef(int32_t)        const;
  void              setUse(Opnd*, int32_t);
  void              setDef(Opnd*, int32_t);
  void              setPredicate(Opnd*, bool=false);

  void              setLiveState(bool live) { _live = live; }
  bool              isLive() const { return _live; }

  void            setCallName(const char*, const char* = NULL, const char* = NULL); //because used only in one place, when call is created

  void            addComment(arena_string& comm)  { comment += comm + "\t"; }
  void            addComment(std::string& comm);
  void            addComment(const char* comm);

  enum CatchSignal { sigill_pc, sigsegv_pc, call_pc = sigsegv_pc };
  void            setCatchType(CatchSignal sig);
  uint64_t        catchPc()              const;
  Block*          getHandlerBlock()      const;
  BlockList       getAllHandlerBlocks()  const;

  uint8_t         mas()                  const { assert_rvm(isMemRead() || isMemWrite() || isMova(), "verify"); return flags >> 24; }
  void            setMas(uint8_t m)            { assert_rvm(isMemRead() || isMemWrite() || isMova(), "verify"); flags |= ((uint32_t)m << 24); }
  uint8_t         getSpecialFlag() const       { assert_rvm(!isMemRead() && !isMemWrite(), "use mas instead"); return flags >> 24; }
  void            setSpecialFlag(uint8_t m)    { assert_rvm(!isMemRead() && !isMemWrite(), "use mas instead"); flags |= ((uint32_t)m << 24); }

public:
  int32_t         id;
  InstCode        opcode;
  OpndVector      uses;
  OpndVector      touch_uses; // touches for safepoints.
  OpndVector      cond_reads;
  uint32_t        flags;      // IF_SPECULATIVE IF_POSSIBLE_EXC IF_USE1_INVERT IF_USE2_INVERT IF_PRED_INVERSE
  int32_t         als_id;
  int32_t         use_clp_num;
  InstCode        linked_opcode;
  uint64_t        pc;         // address of this instruction, at begin id of host block
  uint64_t        next_pc;    // address of the next command
  int32_t         bundle_cycle;
  MemoryInfo*     mem_info;   // high-level information for memory disam

  arena_string    call_name;
  arena_string    comment;

  // TODOII: maybe it is better to create new class for INST_INTRIN
  address         entry;
  size_t          size;

private:
  OpndVector      _defs;
  // TODO: this field is used only in ssa_dce, remove this
  //       implement needed functionality in the pass itself
  bool            _live;
};

//----------------------------------------------------------------------------//

class Bundle : public TangoObj {
public:
  Bundle() : als_available(EU_ALL), lts_size(0), lts16_num(0),
    lts_spec_num(0), len(0), flags(0U) {}
  bool        addInst(Inst*);
  bool        canAddInst(Inst*, int32_t);
  bool        addInst(Inst*, int32_t);
  bool        addMova(Inst*);
  bool        checkLts(int16_t, int16_t, int16_t);
  int32_t     getTotalSize(Inst*, uint32_t, int16_t, int16_t) const;
  bool        containsPrDef() const { return (flags & BF_CONTAINS_PR_DEF); }
  bool        containsCall()  const { return (flags & BF_CONTAINS_CALL); }
  bool        containsSP()    const { return (flags & BF_CONTAINS_SP); }
  bool        containsPE()    const { return (flags & BF_CONTAINS_PE); }
  bool        containsMAS()   const { return (flags & BF_CONTAINS_MAS); }
  bool        containsLM()    const { return (flags & BF_CONTAINS_LOOPM); }
  void        setContainsPrDef()    { flags |= BF_CONTAINS_PR_DEF; }
  void        setContainsCall()     { flags |= BF_CONTAINS_CALL; }
  void        setContainsSP()       { flags |= BF_CONTAINS_SP; }
  void        setContainsPE()       { flags |= BF_CONTAINS_PE; }
  void        setContainsMAS()      { flags |= BF_CONTAINS_MAS; }
  void        setContainsLM()       { flags |= BF_CONTAINS_LOOPM; }
  int16_t     getLtsInstSize(const Inst*, int16_t&, int16_t&);
  int32_t     getExecUnit(const Inst *, uint32_t) const;
  int32_t     getExecUnitSt(uint32_t, uint8_t) const;
  int32_t     getExecUnitLdMas(uint8_t, bool) const;
  int32_t     getExecUnitStMas(uint8_t) const;
  int32_t     getExecUnitLd(uint32_t, uint8_t, bool) const;
  int32_t     toAlsId(uint32_t) const;
  int32_t     getSize() const { return insts.size(); }
  bool        hasOpndDef(Opnd* op) const;
  bool        containsSs(const Inst* inst) const;

public:
  InstList      insts;         // insts belonging to this bundle
  uint32_t      als_available; // exec units available in this bundle
  uint16_t      lts_size;
  uint16_t      lts16_num;
  uint16_t      lts_spec_num;
  uint16_t      len;
  uint16_t      flags;

private:
  uint32_t      getStMask() const { // mask of alus where st operations are placed
    uint32_t st_mask = 0;
    for (InstList::const_iterator inst_it = insts.begin(); inst_it != insts.end(); ++inst_it) {
      const Inst* inst = *inst_it;
      if (inst->isMemWrite()) {
        st_mask |= (1 << inst->als_id);
      }
    }
    return st_mask;
  }
  uint32_t      getLdMask() const { // mask of alus where st operations are placed
    uint32_t ld_mask = 0;
    for (InstList::const_iterator inst_it = insts.begin(); inst_it != insts.end(); ++inst_it) {
      const Inst* inst = *inst_it;
      if (inst->isMemRead()) {
        ld_mask |= (1 << inst->als_id);
      }
    }
    return ld_mask;
  }
  // Returns mask of availiable alu with ids more than last mem operation
  uint32_t      getMemMask() const {
    uint32_t mem_mask = EU_AL_0235;
    for (InstList::const_iterator inst_it = insts.begin(); inst_it != insts.end(); ++inst_it) {
      const Inst* inst = *inst_it;
      if (inst->isMemRead() || inst->isMemWrite()) {
        uint32_t exec_unit = (1 << inst->als_id);
        mem_mask &= ~(exec_unit | (exec_unit - 1)); // forbid all eu smaller then this ld
      }
    }
    return mem_mask;
  }
  
  bool          containsAle(const Inst* inst) const;
  int32_t       cdsSyllablesCount(const Inst* inst) const;
};

//----------------------------------------------------------------------------//

class LiveSet : public TangoObj {
public:
  void    resize(int32_t, int32_t, int32_t, int32_t, int32_t);
  void    resizeAndReset(int32_t, int32_t, int32_t, int32_t, int32_t);
  void    set(const RegOpnd *, bool);
  void    setLive(const RegOpnd* reg);
  void    setDead(const RegOpnd* reg);
  bool    get(const RegOpnd *) const;
  void    operator|=(const RegOpnd*);
  void    operator|=(const LiveSet&);
  bool    operator==(const LiveSet&) const;
  void    operator=(const LiveSet&);
  void    reset();

public:
  BitMap  mask_r;
  BitMap  mask_p;
  BitMap  mask_c;
  BitMap  mask_aa;
  BitMap  mask_s;
};

//----------------------------------------------------------------------------//

class PrSet {
  using PrMask = arena_map<RegOpnd*, BitMap>::type;

public:
  void      clear();
  void      add(RegOpnd*, int32_t);
  void      remove(RegOpnd*);
  
  bool      find(RegOpnd*);
  bool      check(RegOpnd*); 

  BitMap&   getMask(RegOpnd*);
  void      operator |= (const PrSet&);
  
private:
  void set_union(PrMask&, const std::pair<RegOpnd* const, BitMap>&);

public:
  PrMask pr_r;
  PrMask pr_p;
  PrMask pr_c;
  PrMask pr_aa;
  PrMask pr_s;
};

//----------------------------------------------------------------------------//

class Reduction : public TangoObj {
public:
    Reduction(Opnd* reg, Opnd* red_opn, InstCode opcode, Inst* inst) : reduct_reg(reg), last_red_inst(true),
      red_inst(inst), red_opnd(red_opn), is_invar_reduct(true), opcode(opcode) {}
public:
    Opnd*       reduct_reg;  // reduction registr
    Opnd*       red_opnd;    // reduction operand (s = s +  red_opnd)
    InstCode    opcode;      // reduction code
    Inst*       red_inst;   //  reduction's statement
    OpndList    opndstk;    // work field for generation reduction stmt
    bool        last_red_inst;   //  reduction is last in loop for given 'reg'
    bool        is_invar_reduct;  //  reduction operand is invariant
};

typedef arena_list<Reduction*>::type             ReductionList;

//----------------------------------------------------------------------------//

class MemLocalization : public TangoObj {
public:
    MemLocalization(Opnd* reg, Inst* inst) : reg(reg), store(inst), reg_is_live(false), uniq(true)  {}
public:
    Opnd*       reg;  // register for memory
    Inst*       store;     // store instruction
    bool        reg_is_live;     // register is live after loop
    bool        uniq;     // register for memory is the same for all stores
};

typedef arena_list<MemLocalization*>::type             MemLocalizationList;


//----------------------------------------------------------------------------//

class Loop : public TangoObj {
public:
  Loop(int32_t id) : id(id), parent(NULL), child(NULL), sibling(NULL), head(NULL), control_block(NULL),
    tail(NULL), iv(NULL), back_control(NULL), stride(0), _unroll_count(1), _trip_count(-1), _loop_flags(0),
    _is_optimizable(false), _cannot_be_pipelined(true), _has_unlinear_part(false), _gen_dynamic_check(false),
    _no_bases_intersections(false), _ivr_outside_indexes(false), _back_control_in_notail(false),
    _nonstandardloop(false) {}

public:
  // Names for flag bitfields
  enum { Normal=0, Pre=1, Main=2, Post=3, PreMainPostFlagsMask=3,
         MainHasNoPreLoop=4,
         HasExactTripCount=8,
         InnerLoop=16,
         PartialPeelLoop=32,
         PartialPeelFailed=64,
         IsCountedLoop=128,
         IsMultiBlockLoop=128,
         CanBeUnroll=256,
         IsMultiExitLoop=512
      };
  int32_t   id;                // unique loop id
  Loop*     parent;            // parent loop
  Loop*     child;             // child loop
  Loop*     sibling;           // next loop
  Block*    head;              // head block of the loop
  Block*    tail;              // tail block of the loop
  Block*    control_block;     // back control block
  Inst*     iv;                // induction variable
  Inst*     back_control;      // check for back branch
  int32_t   stride;            // loop stride
  ReductionList reds;          // list of reductions
  BlockList back_edge_blocks;  // list if bach edges blocks
  BlockList blocks;            // list of blocks in loop body
  BlockList exit_blocks;       // list of exit blocks for loop
  InstList  arr_access_iters;  // list of array access iterators
  Pragmas   pragmas;           // Pragmas associated with the loop

  int is_inner_loop() const { return child == nullptr; }
  int is_partial_peel_loop() const { return _loop_flags & PartialPeelLoop; }
  int is_normal_loop() const { return (_loop_flags&PreMainPostFlagsMask) == Normal; }
  int is_pre_loop   () const { return (_loop_flags&PreMainPostFlagsMask) == Pre;    }
  int is_main_loop  () const { return (_loop_flags&PreMainPostFlagsMask) == Main;   }
  int is_post_loop  () const { return (_loop_flags&PreMainPostFlagsMask) == Post;   }
  int is_main_no_pre_loop() const { return _loop_flags & MainHasNoPreLoop; }
  bool optimizable() const   { return _is_optimizable; }
  bool noBasesIntersections() const { return _no_bases_intersections; }
  bool genDynamicCheck() const { return _gen_dynamic_check; }
  bool cannotBePipelined() const { return _cannot_be_pipelined; }
  bool HasUnlinearPart() const { return _has_unlinear_part; }
  int partial_peel_has_failed() const { return _loop_flags & PartialPeelFailed; }
  bool has_exact_trip_count() const { return (_loop_flags & HasExactTripCount) != 0; }
  int unroll_count() const { return _unroll_count; }
  float profile_trip_cnt() const { return _profile_trip_cnt; }
  void set_unroll_count(int val) { _unroll_count = val; }
  void set_profile_trip_cnt(float val) { _profile_trip_cnt = val; }
  void add_body_size() {_body_size++;}
  int body_size() {return _body_size;}
  int set_body_size(int val) { _body_size = val;}
  bool is_counted_loop() const { return (_loop_flags & IsCountedLoop) != 0; }
  bool canbeunrolled() const { return (_loop_flags & CanBeUnroll) != 0; }
  void set_is_counted_loop() { _loop_flags |= IsCountedLoop; }
  bool is_multiblocks() const { return (_loop_flags & IsMultiBlockLoop) != 0; }
  void set_is_multiblocks() { _loop_flags |= IsMultiBlockLoop; }
  void clean_is_multiblocks() { _loop_flags &= ~IsMultiBlockLoop; }
  bool is_multiexits() const { return (_loop_flags & IsMultiExitLoop) != 0; }
  void set_is_multiexits() { _loop_flags |= IsMultiExitLoop; }
  void clean_is_multiexits() { _loop_flags &= ~IsMultiExitLoop; }
  void set_is_post_loop() { _loop_flags |= Post; }
  void setOptimizable(bool is_optimizable) { _is_optimizable = is_optimizable; }
  void setCannotBePipelined(bool cannot_be_pipelined) { _cannot_be_pipelined = cannot_be_pipelined; }
  void setHasUnlinearPart(bool has_unlinear_part) { _has_unlinear_part = has_unlinear_part; }
  bool setNoBasesIntersection(bool no_bases_intersections) { _no_bases_intersections = no_bases_intersections; }
  bool setgenDynamicCheck(bool gen_dynamic_check) { _gen_dynamic_check = gen_dynamic_check; }
  bool setIvrOutside(bool ivr_outside_indexes) { _ivr_outside_indexes = ivr_outside_indexes; }
  bool isIvrOutside() const { return _ivr_outside_indexes; }
  bool setBackControlIsNotTail(bool back_control_in_notail) { _back_control_in_notail = back_control_in_notail; }
  bool isBackControlIsNotTail() const { return _back_control_in_notail; }
  bool setNonStandardloop(bool nonstandardloop) { _nonstandardloop = nonstandardloop; }
  bool isNonStandardloop() const { return _nonstandardloop; }
  BlockList   getBlocks() const;  // Calculates list of all blocks in this loop
  RegOpnd*    getRegIV() const {if (!iv) return NULL; Opnd *opn = iv->getDef(0); return opn->asReg();}
  Opnd*       getFinal() const {if (!back_control) return NULL; return back_control->getUse(2); }
  int trip_count() {
     if (has_exact_trip_count()) return _trip_count;
     return -1;
  }
  void set_exact_trip_count(int tc) {
    _trip_count = tc;
    _loop_flags |= HasExactTripCount;
  }
  void unsetset_exact_trip_count() {
    _trip_count = -1;
    _loop_flags &= ~HasExactTripCount;
  }
  void set_loop_flags(short f) { _loop_flags = f; }
  short get_loop_flags() const { return _loop_flags; }
  void IvarCalculate();  // induction variable recognition for loop tree
  void LoopIvpEval();
  bool ivFound(BlockList::iterator bit);
  void ivPropagation(InstList& insts, InstList::iterator init, Opnd *ivopnd, Opnd *useopnd);
  void EvalTailBlock();  // tail block evaluation
  void evalProfileTrip();
  bool createBodyBlockList();
  bool canCopy();
  bool BlockIsAchievable(Block *b1, Block *b2);
  Reduction* getReduction(Inst *inst);
  bool IsInduction(Inst *inst);

private:
  float _profile_trip_cnt;
  int   _unroll_count;      // unroll count
  int   _trip_count;
  int   _body_size;
  short _loop_flags;
  bool  _is_optimizable;
  bool  _cannot_be_pipelined;
  bool  _has_unlinear_part;
  bool  _gen_dynamic_check;
  bool  _no_bases_intersections; // filled during dynchecks generation, used in scheduler
  bool  _ivr_outside_indexes;  // induction variable is used outside array indexes
  bool  _back_control_in_notail;  // back control is in not tail block
  bool  _nonstandardloop;  // non stadard loop (while, no ivp, etc)
};

//----------------------------------------------------------------------------//
class Block : public TangoObj {
public:
  Block(int32_t id, const arena_string& name = "") : id(id), rpo(-1),
    layout_succ(NULL), layout_pred(NULL), parent_l(NULL),
    count_(0), visited(false), postvisited(false), conditional(false),
    freq_(0.0f), name(name), last_phi_inst(NULL), flags(0), region(0) {}

  Block*      pushFront(Inst* inst)     { inst->pc = id; insts.push_front(inst); return this; }
  Block*      pushBack(Inst* inst)      { inst->pc = id; insts.push_back(inst);  return this; }
  Block*      insert(InstList::iterator &, Inst *);
  void        pushFront(Bundle* bundle) { bundles.push_front(bundle); }
  void        pushBack(Bundle* bundle)  { bundles.push_back(bundle); }
  Block*      addPred(Block*);
  Block*      addSucc(Block*);
  void        delPred(Block*);
  void        delSucc(Block*);
  void        clearSuccs();
  void        delInst(Inst*);
  void        setLayout(Block*);

  bool        isTry()            const { return (flags & BB_TRY & ~BB_NO_MERGE) != 0; }
  bool        isCatch()          const { return (flags & BB_CATCH & ~BB_NO_MERGE) != 0; }
  bool        isNoMerge()        const { return (flags & BB_NO_MERGE) != 0; }
  bool        hasImplicitSuccs() const { return (flags & BB_IMPLICIT_SUCCS) != 0; }
  void        setTry()                 { flags |= BB_TRY; }
  void        setCatch()               { flags |= BB_CATCH; }
  void        setImplicitSuccs()       { flags |= BB_IMPLICIT_SUCCS; }
  void        setNoMerge()             { flags |= BB_NO_MERGE; }

  void        setRegion(uint32_t reg);

  void        resetLoopInfo(Block* succ);

  bool        dominates(Block *) const;

  int32_t     count()                const  { return count_; }
  void        setCount(int32_t c)           { count_ = c; }
  void        addCount(int32_t c)           { count_ += c; }

  float       freq()                 const  { return freq_; }
  void        setFreq(float c)              { freq_ = c; }
  void        setParentLoop(Loop* l)        { parent_l = l; }
  bool        canCopy();
  Loop*       parentLoop()           const  { return parent_l; }
public:
  void        delElement(BlockList&, Block*);

  int32_t     id;             // unique block id
  int32_t     rpo;            // reverse post order number
  InstList    insts;          // insts of this block
  BundleList  bundles;        // insts of this block after instruction scheduling

  LiveSet     live_in_set;    // opnds alive at the beginning of a block
  LiveSet     live_out_set;   // opnds alive at the end of a block
  BlockList   preds;          // blocks pedecessing this one in cfg
  BlockList   succs;          // blocks successing this one in cfg
  Block*      layout_succ;    // block successing this one in emitted code
  Block*      layout_pred;    // block pedecessing this one in emitted code
  Loop*       parent_l;       // parent loop for the block
  uint64_t    addr;           // block address in code buffer
  int32_t     count_;         // how many times block executed
  float       freq_;          // execution frequency
  bool        visited;        // field used in block iteration (see LiveAnalyser)
  bool        postvisited;    // field used for loop building (Tarjan algorithm)
  bool        conditional;    // set for conditional blocks inside loop
  arena_string name;
  uint32_t    flags;
  uint32_t    region;

  PrSet       pr_in_set;      // predicate mask for opnds at rhe begining of a block
  PrSet       pr_out_set;     // predicate mask for opnds at the end of a block

  Inst*       last_phi_inst;  // last phi inst whose move was been added to block
};

//----------------------------------------------------------------------------//

class PrGraph {
  struct PrNode {
    int32_t pred;
    int32_t height;
    arena_vector<int32_t>::type succs;
  };

  struct DSNode {
    arena_vector<int32_t>::type set;
    int32_t pr;
  };

public:
  PrGraph() {}

  bool build(Ir*);    
  void updatePrMask(RegOpnd*, bool, BitMap&, bool);
  bool prFound() const {return pr_def_found;}

private:
  void buildByBlock(Block*);
  void updatePrMask(int32_t, BitMap&, bool);
  void updateGraph(Inst*);
  void checkPrDef(Inst*);

  int32_t lca(int32_t, int32_t);
  int32_t findEqual(int32_t);
  void setEqual(int32_t, int32_t);

private:
  arena_vector<PrNode>::type nodes;
  arena_vector<DSNode>::type equal;
  int32_t pr_count;
  int32_t px;
  bool    pr_def_found;
};

//----------------------------------------------------------------------------//

class Ir : public TangoObj {
public:
  Ir(Allocator *alloc);
  BlockList&      search();       // builds post ordered block list
  void            search(Block*, BlockSet&);

  // allocation support
  Allocator*      allocator() const { return _allocator; }

  // ir building support
  ImmOpnd*        newImmOpnd(DataType, uint64_t, uint32_t=0);
  ImmOpnd*        newFImmOpnd(DataType, double, uint32_t=0);
  LabelOpnd*      newLabelOpnd(Block*, float=0.0f);
  SymbolOpnd*     newSymbolOpnd(std::string);
  RegOpnd*        newRegOpnd(RegType, DataType, int32_t=UNALLOCATED, uint32_t=0);
  RegOpnd*        newRegOpnd(Opnd*, int32_t=UNALLOCATED);
  RegOpnd*        newInArg(DataType, int32_t, uint32_t=0);
  RegOpnd*        newOutArg(DataType, int32_t, uint32_t=0);
  RegOpnd*        getLocationReg(RegOpnd*);
  ImmOpnd*        newFapbImm(bool ct, uint8_t dcd, uint8_t fmt, uint8_t mrng,
                             uint8_t d, uint8_t incr, uint8_t ind, uint8_t asz,
                             uint8_t abs, int32_t disp);
  Opnd*           newCtpr() { return newRegOpnd(REG_TYPE_CTPR, DATA_U64, UNALLOCATED); }
  Inst*           newInst(InstCode);
  Bundle*         newBundle();
  Block*          newBlock(const arena_string & = "");
  Block*          newBlock(Block* original, const arena_string& = "");
  Loop*           newLoop();

  void            copyLoopInfo(Loop* l, Block* head_block, Block* tail_block,
                               Loop* new_loop = nullptr);

  // support for bitset+vector based live set manipulation algorithms
  void            initDepVectors();       // reserve space for opnds_x vectors
  void            initDepSets();          // reserve space for RegOpnd::dep_set
  void            initDepIds();           // iterate alive opnds and assign RegOpnd::dep_id
  void            initDepIdInts(Inst* inst);
  void            initDepId(RegOpnd*);    // assign RegOpnd::dep_id
  void            initPrMasks();
  void            resetDepInfo();         // clears opnds_x and dep_id_x
  RegOpndVector&  getDepOpnds(Opnd*);     // returns opnds_x vector for opnds of RegType

  bool            dominates(Block*, Block *);
  bool            regionDominates(Block*, Block*);
  bool            regionDominates(uint32_t, Block*);

  bool            isOSR() const    { return _osr; }
  void            setOSR(bool osr) { _osr = osr; }

  int32_t         lastRelocIndex() { return max_reloc_index - 1; }
  int32_t         calcUnextendedSpillSlotOffset(int idx) const;

  CPUInfo*        cpuinfo() const { return _cpuinfo; }

  bool            failed() const { return fail_reason != NoFail; }
  Inst*           FinalLoopValEvaluation(Loop *l);
  void            clearVisited() {
     for (BlockList::iterator it=blocks.end(); it!=blocks.begin();) {
        Block* block = *(--it);
        block->visited = false;
        block->postvisited = false;
     }
  }

  bool            readGlobalFlags();
  
public:
  Block*          enter;             // cfg enter node
  Loop*           loop_tree;         // top loop
  BlockList       blocks;            // list of cfg blocks (filled by Ir::search())

  RegStack        reg_stack;         // current state of register stack
  MemStack        mem_stack;         // current state of memory stack

  RegOpnd*        stack_pointer;     // register assigned to be memory stack pointer
  RegOpnd*        px;                // fake pred register (alwayas true)
  RegOpnd*        pr_tmp;            // pre-allocated tmp pred register for a single bundle
  ImmOpnd*        wbs;               // wbs value for call insts
  ImmOpnd*        wsz;               // wsz value for call insts
  uint32_t        max_breg;
  bool            has_allocated_breg;
  bool            has_call;          // set up in code selector
  bool            has_safepoint;     // set up in code selector

  PrGraph         predicate_graph;
  
  // support for bitset+vector based live set manipulation algorithms (allocator, live analyser)
  RegOpndVector   opnds_r;           // r    opnds indexed by RegOpnd::dep_id
  RegOpndVector   opnds_p;           // pr   opnds indexed by RegOpnd::dep_id
  RegOpndVector   opnds_c;           // ctpr opnds indexed by RegOpnd::dep_id
  RegOpndVector   opnds_aa;          // array prefetching opnds indexed by RegOpnd::dep_id
  RegOpndVector   opnds_s;           // special opnds indexed by RegOpnd::dep_id

  int32_t         dep_id_r;          // max dep_id for r    opnds
  int32_t         dep_id_p;          // max dep_id for pr   opnds
  int32_t         dep_id_c;          // max dep_id for ctpr opnds
  int32_t         dep_id_aa;         // max dep_id for array prefetching opnds
  int32_t         dep_id_s;          // max dep_id for special opnds

  // Logging accessories
  int32_t         max_opnd_id;
  int32_t         max_inst_id;
  int32_t         max_block_id;
  int32_t         max_loop_id;

  int32_t         max_spill;
  int32_t         max_stack;

  // Relocations
  AddressList     const_relocs;
  DebugInfoTransfer* debug_info;

  // Special registers.
  Opnd*           lsr_reg;
  Opnd*           ctpReg3;
  Opnd*           empty_reg;

  int32_t         spill_base_offset;
  int32_t         stack_direction;
  Opnd*           frame_reg;
  int32_t         max_reloc_index;
  uint32_t        fail_reason;
  uint32_t        last_free_pred;

  // Allocation support
  Allocator*      _allocator;

  CPUInfo*        _cpuinfo;

  Pragmas         pragmas;          // Pragmas associated with the whole Ir
  bool            has_loop_pragmas; // Do every loop has its' own pragmas

  bool            _osr;

  arena_vector<uint32_t>::type region_tree;

  // Flags
  bool _SuperblockScheduling         = false;
  bool _TangoInstructionScheduler    = false;
  bool _TangoGreedyScheduler         = false;
  bool _PredicateRegAllocInScheduler = false;
  bool _UseLockCheckDisambiguation   = false;
  bool _UseLandp                     = false;
  bool _TangoSsaDce                  = false;
  bool _TangoUnroll                  = false;
  bool _TangoMoveInvariants          = false;
  bool _AllowCopyForSuperblocks      = false;
  bool _ForbidCreateCompCode         = false;
  bool _UsePredsForStores            = false;
  bool _Verbose                      = false;
  bool _TangoUseCopyPropagation      = false;
  bool _TangoLoopsBuild              = false;
  bool _TangoIfConversion            = false;
  bool _GenerateDynamicChecks        = false;
  bool _TangoMultiBlocksLoop         = false;
  bool _UseMemoryDisambiguation      = false;
  bool _UseSoftwarePipelining        = false;
  bool _OptReduction                 = false;
  bool _UseFloatReduction            = false;
  bool _OptMemoryLocalization        = false;
  bool _OptRotateBaseMoving          = false;
  bool _RAW_SXTProcessing            = false;
  bool _UsePipelineTransformer       = false;
  bool _TreeHeightBalancing          = false;
  bool _TangoColdPathMerging         = false;
  bool _TangoCmpIBranchMacro         = false;
  bool _TangoInsertIBranches         = false;
  bool _TangoFMAMacro                = false;
  bool _PredicateAnalyses            = false;


  int32_t _DisamLdStLimit = -1;
  int32_t _TangoCmpBlockLimit = 40;
};

//----------------------------------------------------------------------------//

} // namespace tango

#endif // SHARE_VM_TANGO_IR_HPP
