#ifdef __cplusplus
#ifndef ASM_E2K_OPERANDS_H
#define ASM_E2K_OPERANDS_H

#include <inttypes.h>
#include <assert.h>
#include "registers.h"
#include "macro.h"

namespace asm_e2k {

typedef int Relocation;
enum {
    RelocationNone = ASM_E2K_RELOCATION_NONE
};

/* -------------------------------------------------------------------*/

class Register;
class Immediate;

class Operand {
protected:
    Operand() {}
    Operand(DataTypes data_type, bool is_reg, uint64_t value) : is_reg(is_reg), data_type(data_type), value(value) {}

public:

    Operand(RegCode reg) : Operand(DATA_U64, true, reg) { }

    Operand &operator= (const Operand& other) {
        this->is_reg = other.is_reg;
        this->data_type = other.data_type;
        this->reloc = other.reloc;
        this->value = other.value;
        return *this;
    }

    Operand &operator= (const RegCode reg) {
        *this = Operand(DATA_U64, true, reg);
        return *this;
    }

    Register& asReg() { ASM_E2K_ASSERT(isReg()); return *(Register*)(this); }
    const Register& asReg() const { ASM_E2K_ASSERT(isReg()); return *(const Register*)(this); }

    Immediate& asImm() { ASM_E2K_ASSERT(isImm()); return *(Immediate*)(this); }
    const Immediate& asImm() const { ASM_E2K_ASSERT(isImm()); return *(const Immediate*)(this); }

    bool isImm() const { return !is_reg; }

    bool isReg() const { return is_reg; }

    bool isBRreg() const { return isReg() && value.reg >= EREG_BR0 && value.reg <= EREG_BR127; }

    bool isGreg() const { return isReg() && value.reg <= EREG_G31; }

    bool isRreg() const { return isReg() && value.reg >= EREG_R0 && value.reg <= EREG_R63; }

    bool isPreg() const { return isReg() && value.reg >= EREG_PR0 && value.reg <= EREG_PRX; }

    bool isCreg() const { return isReg() && value.reg >= EREG_CTPR1 && value.reg <= EREG_CTPR3; }

    uint64_t getValue() const;

    uint64_t getValueOld() const;

    bool      is_reg    : 1;
    DataTypes data_type : 4;
    Relocation reloc    : 27;
    union _value {
        uint64_t  i;
        double    d;
        RegCode reg;
        _value() {}
        _value(RegCode _reg) : reg(_reg) { }
        _value(uint64_t _i) : i(_i) { }
        _value(double _d) : d(_d) { }
    } value;
};

/* -------------------------------------------------------------------*/

class Immediate : public Operand {
public:
    explicit Immediate(uint64_t _value, DataTypes dt = DATA_U64, Relocation kind = RelocationNone) : Operand(dt, false, _value) { this->reloc = kind; }
    explicit Immediate(int32_t  _value, DataTypes dt = DATA_I32, Relocation kind = RelocationNone) : Operand(dt, false, _value) { this->reloc = kind; }

//    template<class FROM> Immediate(FROM from);

    uint64_t getValue() const { return value.i; }

    double getDouble() const { return value.d; }

    void setInt(uint64_t i) { value.i = i; }

    void setDouble(double d) { value.d = d; }

    bool isSigned() const { return data_type % 2 != 0; }

    bool isInt() const { return data_type <= DATA_U64; }

    uint64_t getInt() const { return value.i; }

    int32_t fits_in_size() const;
};

static Immediate imm0(0, DATA_I32);
static Immediate imm1(1, DATA_I32);

/* -------------------------------------------------------------------*/

class Register : public Operand {
public:

    Register(const Register& other) {
        (*this) = other;
    }

    Register &operator= (const Register &right) {
        is_reg = true;
        data_type = right.data_type;
        reloc = right.reloc;
        value.reg = right.value.reg;
        return *this;
    }

    Register &operator= (const RegCode reg_code) {
        *this = Register(reg_code);
        return *this;
    }

    Register(uint32_t reg_code) : Operand(DATA_U64, true, reg_code) { }

    RegCode reg_code() const { return (RegCode) value.reg; }

    bool is_valid() const { return 0 <= reg_code() && reg_code() < EREG_LAST; }
};

} // namespace asm_e2k

#endif // ASM_E2K_OPERANDS_H

#endif // __cplusplus