#ifndef TANGO_TANGO_INST_HPP
#define TANGO_TANGO_INST_HPP

#include <stdint.h>
#include <stdio.h>
#include <iostream>
#include <vector>
#include <list>
#include <set>
#include <unordered_set>
#include <unordered_map>
#include <map>
#include <algorithm>
#include <cstddef>
#include <string>
#include <sstream>
#include <unordered_map>
#include <unordered_set>

#include "tango_allocator.hpp"

namespace tango {

//----------------------------------------------------------------------------//

template<typename T = int>
class arena_allocator {
public:
  using value_type      = T;
  using size_type       = std::size_t;
  using difference_type = std::ptrdiff_t;
  using pointer         = T*;
  using reference       = T&;
  using const_pointer   = const T*;
  using const_reference = const T&;
  //using propagate_on_container_move_assignment = std::false_type;
  //using propagate_on_container_copy_assignment = std::true_type;
  //using propagate_on_container_swap = std::true_type;

public:
  arena_allocator() : _alloc(nullptr) { }
  arena_allocator(TangoObj::Allocator *alloc) : _alloc(alloc) {}
  arena_allocator(const arena_allocator<T>& alloc) : _alloc(alloc._alloc) { }
  template<typename U>
  arena_allocator(const arena_allocator<U>& alloc) : _alloc(alloc._alloc) { }

  static size_type max_size() { return (std::numeric_limits<size_type>::max)(); }

  void construct(const pointer ptr, const T& t) { ::new (ptr) T(t); }
  void destroy(pointer p) {}

  static void deallocate(const pointer, const size_type) {}

  template<typename _Tp1>
  struct rebind {
    typedef arena_allocator<_Tp1> other;
  };

  constexpr bool operator == (const arena_allocator<T>& other) const { return true; }
  constexpr bool operator != (const arena_allocator<T>& other) const { return false; }
  arena_allocator<T>& operator =(const arena_allocator<T>& other) noexcept {
    this->_alloc = other._alloc;
    return *this;
  }

  pointer allocate(const size_type sz) {
    if (!_alloc) {
      _alloc = _shared_allocator;
    }
    assert_rvm(_alloc, "Allocator uninitialized");
    return static_cast<pointer>(_alloc->allocate(sizeof(T) * sz));
  }
  pointer allocate(const size_type sz, const void *hint) { return allocate(sz); }

  TangoObj::Allocator *_alloc;
};

//----------------------------------------------------------------------------//

template<typename T = int>
class persistent_allocator {
public:
  using value_type      = T;
  using size_type       = std::size_t;
  using difference_type = std::ptrdiff_t;
  using pointer         = T*;
  using reference       = T&;
  using const_pointer   = const T*;
  using const_reference = const T&;
  //using propagate_on_container_move_assignment = std::false_type;
  //using propagate_on_container_copy_assignment = std::true_type;
  //using propagate_on_container_swap = std::true_type;

public:
  persistent_allocator() : _alloc(&TangoBufferAllocator::getPersistentAllocator()) { }
  persistent_allocator(PersistentObj::Allocator *alloc) : _alloc(alloc) {}
  persistent_allocator(const persistent_allocator<T>& alloc) : _alloc(alloc._alloc) { }
  template<typename U>
  persistent_allocator(const persistent_allocator<U>& alloc) : _alloc(alloc._alloc) { }

  static size_type max_size() { return (std::numeric_limits<size_type>::max)(); }

  void construct(const pointer ptr, const T& t) {
    ::new (ptr) T(t);
  }
  void destroy(pointer p) {}

  static void deallocate(const pointer, const size_type) {}

  template<typename _Tp1>
  struct rebind {
    typedef persistent_allocator<_Tp1> other;
  };

  constexpr bool operator == (const persistent_allocator<T>& other) const { return true; }
  constexpr bool operator != (const persistent_allocator<T>& other) const { return false; }
  persistent_allocator<T>& operator =(const persistent_allocator<T>& other) noexcept {
    this->_alloc = other._alloc;
    return *this;
  }

  pointer allocate(const size_type sz, const void *hint) { return allocate(sz); }
  pointer allocate(const size_type sz) {
    return static_cast<pointer>(_alloc->allocate(sizeof(T) * sz));
  }

  PersistentObj::Allocator *_alloc;
};

//----------------------------------------------------------------------------//

template<typename U>
static constexpr bool operator == (const arena_allocator<U>& left, const arena_allocator<U>& right) { return true; }

template<typename U>
static constexpr bool operator != (const arena_allocator<U>& left, const arena_allocator<U>& right) { return false; }

template<typename T, template<typename> class Allocator = arena_allocator>
struct arena_vector
{
  typedef Allocator<T> allocator;
  typedef std::vector<T, allocator> type;
};

template<typename T, template<typename> class Allocator = arena_allocator>
struct arena_list
{
  typedef Allocator<T> allocator;
  typedef std::list<T, allocator> type;
};

template<typename T, typename Comparator = std::less<T>, template<typename> class Allocator = arena_allocator >
struct arena_set
{
  typedef Allocator<T> allocator;
  typedef std::set<T, Comparator, allocator> type;
};

template<typename T, typename Hash = std::hash<T>, typename Pred = std::equal_to<T>, template<typename> class Allocator = arena_allocator >
struct arena_unordered_set
{
  typedef Allocator<T> allocator;
  typedef std::unordered_set<T, Hash, Pred, allocator> type;
};

template<typename T, typename Comparator = std::less<T>, template<typename> class Allocator = arena_allocator >
struct arena_multiset
{
  typedef Allocator<T> allocator;
  typedef std::multiset<T, Comparator, allocator> type;
};

template<typename K, typename T, typename Comparator = std::less< K >, template<typename> class Allocator = arena_allocator >
struct arena_map
{
  typedef Allocator<std::pair<const K, T> > allocator;
  typedef std::map<K, T, Comparator, allocator> type;
};

template<typename K, typename T, typename Hash = std::hash<K>, typename Pred = std::equal_to<K>, template<typename> class Allocator = arena_allocator >
struct arena_unordered_map
{
  typedef Allocator<std::pair<const K, T> > allocator;
  typedef std::unordered_map<K, T, Hash, Pred, allocator> type;
};

template<typename K, typename T, typename Comparator = std::less< K >, template<typename> class Allocator = arena_allocator >
struct arena_multimap
{
  typedef Allocator<std::pair<const K, T> > allocator;
  typedef std::multimap<K, T, Comparator, allocator> type;
};

template <typename T, template<typename> class Allocator = arena_allocator>
struct arena_basic_string {
  typedef Allocator<T> allocator;
  typedef std::basic_string< T, std::char_traits<T>, allocator > type;
};

using arena_string    = arena_basic_string<char>::type;
using arena_wstring   = arena_basic_string<wchar_t>::type;
using arena_u16string = arena_basic_string<char16_t>::type;
using arena_u32string = arena_basic_string<char32_t>::type;

using persistent_string = arena_basic_string<char, persistent_allocator>::type;

template <typename T, template<typename> class Allocator = arena_allocator>
struct arena_basic_stringstream {
  typedef Allocator<T> allocator;
  typedef std::basic_stringstream< T, std::char_traits<T>, allocator > type;
};

typedef arena_basic_stringstream<char>::type                       arena_stringstream;
typedef arena_basic_stringstream<char, persistent_allocator>::type persistent_stringstream;

//----------------------------------------------------------------------------//

#define BIT(n) (1U << (n))

#define TANGO_UNREACHABLE  ShouldNotReachHere()

#define REG_NUM_R          96   // number of work registers (32 static + 64 stack)
#define REG_NUM_BR         128  // number of rotated registers
#define REG_NUM_PR         32   // number of pr registers
#define REG_NUM_PRX         1   // number of prx registers
#define REG_NUM_BPR        32   // number of rotating predicates
#define REG_NUM_CTPR        3   // number of ctpr registers
#define REG_NUM_S           3   // number of special registers

//numbers for array prefetching registers
#define REG_NUM_AAD        32
#define REG_NUM_AASTI      16
#define REG_NUM_AAIND      16
#define REG_NUM_AAINCR      8
#define REG_NUM_AALDI      64
#define REG_NUM_AALDA      64
#define REG_NUM_AASPEC      4

#define REG_NUM_AA_TOTAL  204   // number of all aa registers

#define REG_NUM           (136 + 128 + REG_NUM_BPR + REG_NUM_AA_TOTAL)   // number of all registers

#define MAX_REG_ARG         8   // number of in_args resided on registers
#define MEM_ARG_SLOT_SIZE   8   // in/out arg memory slot size in bytes
#define MEM_ALIGNMENT       2   // bytes
#define REG_ALIGNMENT       2   // n regs

#define ILLEGAL_ALS_ID     -1   // calculating als id in Bundle
#define UNALLOCATED        -1   // location value of unallocated RegOpnd
#define REG_PRX      EREG_PRX  // fake pr register
#define REG_PRX_NUM  (EREG_PRX - EREG_PR0)

#define OF_IN_ARG        BIT(0)   // opnd is in arg
#define OF_OUT_ARG       BIT(1)   // opnd is out arg
#define OF_MANAGED_PTR   BIT(2)   // opnd is object address
#define OF_DERIVED_PTR   BIT(3)   // opnd is derived object address
#define OF_NON_ZERO      BIT(4)   // opnd is known to be non-zero value
#define OF_NARROW_PTR    BIT(5)   // opnd is narrow object address
#define OF_EXACT_SIZE    BIT(6)   // Strict exact size in LTS
#define OF_INVERTION     BIT(7)   // Inverted predicate
#define OF_UNSPILLABLE   BIT(8)   // opnd is unspillable, use with caution and only when absolutely necessary
#define OF_VALUE         BIT(9)   // opnd is tagged value

#define IF_SPECULATIVE   BIT(0)   // instuction is speculative
#define IF_POSSIBLE_EXC  BIT(1)   // instruction can throw exception
#define IF_USE1_INVERT   BIT(2)   // use1 inverted (andp, movep)
#define IF_USE2_INVERT   BIT(3)   // use2 inverted (andp, movep)
#define IF_USE3_INVERT    BIT(4)   // use3 inverted (mlandp)
#define IF_USE4_INVERT    BIT(5)   // use4 inverted (mlandp)
#define IF_USE5_INVERT    BIT(6)   // use5 inverted (mlandp)
#define IF_USE6_INVERT    BIT(7)   // use6 inverted (mlandp)
#define IF_PRED_INVERSE   BIT(8)   // instruction's predicate is inverted
#define IF_SAFEPOINT      BIT(9)	  // instruction is a safepoint
#define IF_SPILL          BIT(10)  // instruction was added by spill generator
#define IF_CATCH_PC       BIT(11)  // '0' means next_pc, '1' means pc
#define IF_SINGLE         BIT(12)  // instruction can't be bundled
#define IF_SIDE_EFFECT    BIT(13)  // instruction has side effects and can't be removed
#define IF_UNCHECKPTR_LD  BIT(14)  // instruction is a load with no check for null ptr
#define IF_INVARIANT_LD   BIT(15)  // instruction is a load for array size
#define IF_NO_SIDE_EFFECT BIT(16)  // instruction is a load for array size

#define BF_CONTAINS_PR_DEF  BIT(0)   // bundle contains predicate def
#define BF_CONTAINS_CALL    BIT(1)   // bundle contains call
#define BF_CONTAINS_SP      BIT(2)   // bundle contains safepoint
#define BF_CONTAINS_PE      BIT(3)   // bundle contains possible excpeiton instruction
#define BF_CONTAINS_MAS     BIT(4)   // bundle contains MAS operation
#define BF_CONTAINS_LOOPM   BIT(5)   // bundle contains loop mode

#define BB_NO_MERGE         BIT(0)                // Basic block should not be merged with others
#define BB_TRY              BIT(1) // Basic block is a try
#define BB_CATCH            BIT(2) // Basic block is a catch
#define BB_IMPLICIT_SUCCS   BIT(3) //FIXME::Merge?// Basic block has successors not reachable by branches
#define BB_FORCE_LAYOUT     BIT(4) // Basic block layout successor is set by selector

#define LOOP_IVDEP          BIT(0)
#define LOOP_UNROLL         BIT(1)
#define LOOP_UNCHECK        BIT(2)
#define LOOP_FFAST          BIT(3)
#define LOOP_SIMD           BIT(4)

#define SETWD_LTS_SIZE   4
#define AA_LTS_SIZE      4
#define MAX_LTS_SIZE    16
#define MAX_LTS16_NUM    4
#define MAX_LTS_SPEC_NUM 3
#define MAX_MOVAS_NUM    4
#define TOTAL_BUNDLE_HS 32


#define FIRST_ARG_POS 3

// block layouter
#define BRANCH_TRG 1
#define CT_PRED  0
#define GOTO_TRG 1
#define COND_TRG 2

// encoder
const uint32_t P0 = 0x0;      // internal pred register for clp0
const uint32_t P1 = 0x1;      // internal pred register for clp0
const uint32_t P2 = 0x2;      // internal pred register for clp1
const uint32_t P3 = 0x3;      // internal pred register for clp1
const uint32_t P4 = 0x4;      // internal pred register for clp1
const uint32_t P5 = 0x5;      // internal pred register for clp1

#define SIGNIFICANT_BUNDLES_BETW_SPB 4    //bundles count which we look at when making edges between superblocks

#define REASONABLY_LARGE_VALUE 1e15
#define SP_SPILL_COST 1e20
#define LAYOUT_LOW_FREQ 0.02f

//----------------------------------------------------------------------------//

class Opnd;
class Inst;
class Bundle;
class Block;
class RegOpnd;
class Reloc;

//----------------------------------------------------------------------------//

using DoubleVector       = arena_vector<double>::type;
using OpndSet            = arena_set<Opnd*>::type;
using BlockSet           = arena_set<Block*>::type;
using RegOpndSet         = arena_set<RegOpnd*>::type;
using ConstBlockSet      = arena_set<const Block*>::type;
using OpndList           = arena_list<Opnd*>::type;
using InstList           = arena_list<Inst*>::type;
using BundleList         = arena_list<Bundle*>::type;
using BlockList          = arena_list<Block*>::type;
using ConstBlockList     = arena_list<const Block*>::type;
using RegOpndList        = arena_list<RegOpnd*>::type;
using AddressList        = arena_list<u_char*>::type;
using RelocVector        = arena_vector<Reloc*>::type;
using OpndVector         = arena_vector<Opnd*>::type;
using RegOpndVector      = arena_vector<RegOpnd*>::type;
using InstVector         = arena_vector<Inst*>::type;
using BlockVector        = arena_vector<Block*>::type;
using RegOpnd2RegOpndMap = arena_map<RegOpnd*, RegOpnd*>::type;
using uint642OpndMap     = arena_map<uint64_t, Opnd* >::type;

//----------------------------------------------------------------------------//

enum DataType {
  DATA_U1,
  DATA_I8,
  DATA_U8,
  DATA_I16,
  DATA_U16,
  DATA_I32,
  DATA_U32,
  DATA_I64,
  DATA_U64,
  DATA_S,
  DATA_D,
  DATA_X,
  DATA_VOID,
  DATA_LAST
};

//----------------------------------------------------------------------------//

enum OpndType {
  OPND_IMM,
  OPND_REG,
  OPND_LABEL,
  OPND_SYMBOL,
  OPND_LAST
};

//----------------------------------------------------------------------------//

enum RegType {
  REG_TYPE_R,
  REG_TYPE_PR,
  REG_TYPE_CTPR,
  REG_TYPE_AAD,     // AA registers used for array prefetching
  REG_TYPE_AASTI,
  REG_TYPE_AAIND,
  REG_TYPE_AAINCR,
  REG_TYPE_AALDI,
  REG_TYPE_AALDA,
  REG_TYPE_AASPEC,
  REG_TYPE_SPECIAL, // lsr + emty regs
  REG_TYPE_LAST
};

//----------------------------------------------------------------------------//

enum InstCode {
  INST_FIRST,

  INST_GETSP = INST_FIRST,
  INST_SETWD,   // also sets setbn

  INST_AND,
  INST_ANDN,    // a & ~b
  INST_OR,
  INST_ORN,     // a | ~b
  INST_XOR,
  INST_XORN,    // a ^ ~b

  INST_SHL,
  INST_SHR,
  INST_SCL,
  INST_SCR,
  INST_SAR,

  INST_GETF,
  INST_INSF,

  INST_SXT,
  INST_ZXT,
  INST_RAW_SXT,
  INST_SUB,
  INST_ADD,
  INST_MERGE,     // (merge pred, a, b, res) => (res = pred ? a : b)

  // ANDP and LANDP: res.v = ((nega ^ a.v) && (negb ^ b.v))
  INST_ANDP,      // res.t = (a.tag || b.tag)
  INST_LANDP,     // res.t = (a.tag || ((nega ^ a.v) && b.tag))
  INST_MLANDP,    // res.t = (a.tag || ((nega ^ a.v) && b.tag))
  INST_MOVEP,

  INST_SHL_ADD,
  INST_SHL_AND,
  INST_SHR_AND,

  INST_MUL,
  INST_SMULX,
  INST_UMULX,
  INST_UDIV,
  INST_SDIV,

  INST_ADDS_FB,
  INST_ADDS_FH,
  INST_ADDS_FW,
  INST_ADDD_FD,
  INST_SUBS_FB,
  INST_SUBS_FH,
  INST_SUBS_FW,
  INST_SUBD_FD,
  INST_UMULX_FB,
  INST_UMULX_FH,
  INST_UMULX_FW,
  INST_SMULX_FB,
  INST_SMULX_FH,
  INST_SMULX_FW,

  INST_FSUB,
  INST_FADD,
  INST_FXADD,
  INST_FMUL,
  INST_FXMULX,
  INST_FXMULD,
  INST_FDIV,

  INST_FADD_FADD,
  INST_FSUB_FADD,
  INST_FMUL_FADD,

  INST_FADD_FSUB,
  INST_FSUB_FSUB,
  INST_FMUL_FSUB,

  INST_FADD_FRSUB,
  INST_FSUB_FRSUB,
  INST_FMUL_FRSUB,

  INST_FMIN,
  INST_FMAX,

  INST_FRCP,      // 1/x (single)

  INST_FSQRTI,    // 1st part of double SQRT
  INST_FSQRTT,    // 2nd part of double SQRT

  INST_FSQRT,     // single SQRT

  INST_FRSQRT,    // x^(-1/2) (single)

  INST_LDB,       // all ld insts are grouped together
  INST_LDH,
  INST_LDW,
  INST_LDD,

  INST_STB,       // all st insts are grouped together
  INST_STH,
  INST_STW,
  INST_STD,
  INST_UNLOCK_ST, // it is branch and st in one

  INST_WAIT,

  // convert integer flag to logical predicate
  // +-----+-------+----+-----+----+----+---+----+---+----+---+----+
  // | tag |  "0"  | of | "0" | sf | zf |"0"| af |"0"| pf |"0"| cf |
  // +-----+-------+----+-----+----+----+---+----+---+----+---+----+
  //  t1-t0 31---12  11  10--8  7    6    5   4    3   2    1   0
  //
  INST_CCTOPO,    //               || overflow flag(OF)
  INST_CCTOPB,    // "< " unsigned || carry flag(CF)
  INST_CCTOPE,    // "=="          || zero flag(ZF)
  INST_CCTOPBE,   // "<=" unsigned || (CF | ZF)
  INST_CCTOPS,    // negative      || sign flag(SF)
  INST_CCTOPP,    // odd           || parity flag(PF)
  INST_CCTOPL,    // "< " signed   || (SF ^ OF)
  INST_CCTOPLE,   // "<=" signed   || (SF ^ OF | ZF)

  // predicate - result
  INST_CMPO,      // overflow
  INST_CMPB,      // "< " unsigned
  INST_CMPE,      // "=="
  INST_CMPBE,     // "<=" unsigned
  INST_CMPS,      // negative
  INST_CMPP,      // odd
  INST_CMPL,      // "< " signed
  INST_CMPLE,     // "<=" signed

  INST_CMPANDE,   // zero
  INST_CMPANDS,   // negative
  INST_CMPANDP,   // odd
  INST_CMPANDLE,  // "<=" signed

  INST_FCMPEQ,
  INST_FCMPLT,
  INST_FCMPLE,
  INST_FCMPUOD,
  INST_FCMPNEQ,
  INST_FCMPNLT,     // >=
  INST_FCMPNLE,     // >
  INST_FCMPOD,

  // mask - result
  INST_FCMPEQ_M,
  INST_FCMPLT_M,
  INST_FCMPLE_M,
  INST_FCMPUOD_M,
  INST_FCMPNEQ_M,
  INST_FCMPNLT_M,
  INST_FCMPNLE_M,
  INST_FCMPOD_M,

  // flag - result
  INST_FCMPOD_F,
  INST_FCMPUD_F,

  INST_ITOF,
  INST_ISTOFS,
  INST_ISTOFD,
  INST_FTOI,
  INST_FTOD,
  INST_FTOX,
  INST_DTOF,
  INST_DTOX,
  INST_MOVIF,

  INST_BITREV,
  INST_LZCNT,
  INST_POPCNT,

  INST_RR,  // read register
  INST_RW,  // write register

  INST_DISP,
  INST_LDISP,
  INST_SDISP, // system disp for syscalls
  INST_MOVTD,
  INST_RETURN,

  INST_BRANCH,
  INST_IBRANCH,
  INST_MCMPIBRANCH,
  INST_CT_COND,
  INST_CT_RET,
  INST_SWITCH,
  INST_EX_SWITCH,
  INST_CALL,

  // TODO: add all packed instructions
  INST_PADDW,
  INST_PADDH,
  INST_PADDB,
  INST_PSUBW,
  INST_PFADD,
  INST_PFSUB,
  INST_PFMUL,
  INST_PFDIV,
  INST_PSHUFB,
  INST_PSHUFH,

  INST_LDAAB,
  INST_LDAAH,
  INST_LDAAW,
  INST_LDAAD,
  INST_LDAAQ,
  INST_AAURR,
  INST_AAURRD,

  INST_STAAB,
  INST_STAAH,
  INST_STAAW,
  INST_STAAD,
  INST_STAAQ,
  INST_AAURW,
  INST_AAURWD,

  INST_MOVAB,
  INST_MOVAH,
  INST_MOVAW,
  INST_MOVAD,

  INST_BAP, // also sets eap
  INST_ABN, // also sets alc, abp, abg

  INST_MOV,
  INST_MOVIBRANCH,
  INST_NOP,
  INST_PHI,
  INST_CMP_XCHG, // TODO remove

  INST_ROOT,

  INST_INTRIN,

  INST_LAST
};

//----------------------------------------------------------------------------//

enum RegCode {
  EREG_G0,  EREG_G1,  EREG_G2,  EREG_G3,  EREG_G4,  EREG_G5,  EREG_G6,  EREG_G7,  EREG_G8,  EREG_G9,
  EREG_G10, EREG_G11, EREG_G12, EREG_G13, EREG_G14, EREG_G15, EREG_G16, EREG_G17, EREG_G18, EREG_G19,
  EREG_G20, EREG_G21, EREG_G22, EREG_G23, EREG_G24, EREG_G25, EREG_G26, EREG_G27, EREG_G28, EREG_G29,
  EREG_G30, EREG_G31,

  EREG_R0,  EREG_R1,  EREG_R2,  EREG_R3,  EREG_R4,  EREG_R5,  EREG_R6,  EREG_R7,  EREG_R8,  EREG_R9,
  EREG_R10, EREG_R11, EREG_R12, EREG_R13, EREG_R14, EREG_R15, EREG_R16, EREG_R17, EREG_R18, EREG_R19,
  EREG_R20, EREG_R21, EREG_R22, EREG_R23, EREG_R24, EREG_R25, EREG_R26, EREG_R27, EREG_R28, EREG_R29,
  EREG_R30, EREG_R31, EREG_R32, EREG_R33, EREG_R34, EREG_R35, EREG_R36, EREG_R37, EREG_R38, EREG_R39,
  EREG_R40, EREG_R41, EREG_R42, EREG_R43, EREG_R44, EREG_R45, EREG_R46, EREG_R47, EREG_R48, EREG_R49,
  EREG_R50, EREG_R51, EREG_R52, EREG_R53, EREG_R54, EREG_R55, EREG_R56, EREG_R57, EREG_R58, EREG_R59,
  EREG_R60, EREG_R61, EREG_R62, EREG_R63,

  EREG_BR0,   EREG_BR1,   EREG_BR2,   EREG_BR3,   EREG_BR4,   EREG_BR5,   EREG_BR6,   EREG_BR7,   EREG_BR8,   EREG_BR9,
  EREG_BR10,  EREG_BR11,  EREG_BR12,  EREG_BR13,  EREG_BR14,  EREG_BR15,  EREG_BR16,  EREG_BR17,  EREG_BR18,  EREG_BR19,
  EREG_BR20,  EREG_BR21,  EREG_BR22,  EREG_BR23,  EREG_BR24,  EREG_BR25,  EREG_BR26,  EREG_BR27,  EREG_BR28,  EREG_BR29,
  EREG_BR30,  EREG_BR31,  EREG_BR32,  EREG_BR33,  EREG_BR34,  EREG_BR35,  EREG_BR36,  EREG_BR37,  EREG_BR38,  EREG_BR39,
  EREG_BR40,  EREG_BR41,  EREG_BR42,  EREG_BR43,  EREG_BR44,  EREG_BR45,  EREG_BR46,  EREG_BR47,  EREG_BR48,  EREG_BR49,
  EREG_BR50,  EREG_BR51,  EREG_BR52,  EREG_BR53,  EREG_BR54,  EREG_BR55,  EREG_BR56,  EREG_BR57,  EREG_BR58,  EREG_BR59,
  EREG_BR60,  EREG_BR61,  EREG_BR62,  EREG_BR63,  EREG_BR64,  EREG_BR65,  EREG_BR66,  EREG_BR67,  EREG_BR68,  EREG_BR69,
  EREG_BR70,  EREG_BR71,  EREG_BR72,  EREG_BR73,  EREG_BR74,  EREG_BR75,  EREG_BR76,  EREG_BR77,  EREG_BR78,  EREG_BR79,
  EREG_BR80,  EREG_BR81,  EREG_BR82,  EREG_BR83,  EREG_BR84,  EREG_BR85,  EREG_BR86,  EREG_BR87,  EREG_BR88,  EREG_BR89,
  EREG_BR90,  EREG_BR91,  EREG_BR92,  EREG_BR93,  EREG_BR94,  EREG_BR95,  EREG_BR96,  EREG_BR97,  EREG_BR98,  EREG_BR99,
  EREG_BR100, EREG_BR101, EREG_BR102, EREG_BR103, EREG_BR104, EREG_BR105, EREG_BR106, EREG_BR107, EREG_BR108, EREG_BR109,
  EREG_BR110, EREG_BR111, EREG_BR112, EREG_BR113, EREG_BR114, EREG_BR115, EREG_BR116, EREG_BR117, EREG_BR118, EREG_BR119,
  EREG_BR120, EREG_BR121, EREG_BR122, EREG_BR123, EREG_BR124, EREG_BR125, EREG_BR126, EREG_BR127,

  EREG_PR0,  EREG_PR1,  EREG_PR2,  EREG_PR3,  EREG_PR4,  EREG_PR5,  EREG_PR6,  EREG_PR7,  EREG_PR8,  EREG_PR9,
  EREG_PR10, EREG_PR11, EREG_PR12, EREG_PR13, EREG_PR14, EREG_PR15, EREG_PR16, EREG_PR17, EREG_PR18, EREG_PR19,
  EREG_PR20, EREG_PR21, EREG_PR22, EREG_PR23, EREG_PR24, EREG_PR25, EREG_PR26, EREG_PR27, EREG_PR28, EREG_PR29,
  EREG_PR30, EREG_PR31, EREG_PRX,

  EREG_BPR0,  EREG_BPR1,  EREG_BPR2,  EREG_BPR3,  EREG_BPR4,  EREG_BPR5,  EREG_BPR6,  EREG_BPR7,  EREG_BPR8,  EREG_BPR9,
  EREG_BPR10, EREG_BPR11, EREG_BPR12, EREG_BPR13, EREG_BPR14, EREG_BPR15, EREG_BPR16, EREG_BPR17, EREG_BPR18, EREG_BPR19,
  EREG_BPR20, EREG_BPR21, EREG_BPR22, EREG_BPR23, EREG_BPR24, EREG_BPR25, EREG_BPR26, EREG_BPR27, EREG_BPR28, EREG_BPR29,
  EREG_BPR30, EREG_BPR31,

  EREG_CTPR1, EREG_CTPR2, EREG_CTPR3,

  EREG_AAD0, EREG_AAD1, EREG_AAD2, EREG_AAD3, EREG_AAD4, EREG_AAD5, EREG_AAD6, EREG_AAD7, EREG_AAD8, EREG_AAD9,
  EREG_AAD10, EREG_AAD11, EREG_AAD12, EREG_AAD13, EREG_AAD14, EREG_AAD15, EREG_AAD16, EREG_AAD17, EREG_AAD18, EREG_AAD19,
  EREG_AAD20, EREG_AAD21, EREG_AAD22, EREG_AAD23, EREG_AAD24, EREG_AAD25, EREG_AAD26, EREG_AAD27, EREG_AAD28, EREG_AAD29,
  EREG_AAD30, EREG_AAD31,

  EREG_AASTI0, EREG_AASTI1, EREG_AASTI2, EREG_AASTI3, EREG_AASTI4, EREG_AASTI5, EREG_AASTI6, EREG_AASTI7, EREG_AASTI8, EREG_AASTI9,
  EREG_AASTI10, EREG_AASTI11, EREG_AASTI12, EREG_AASTI13, EREG_AASTI14, EREG_AASTI15,

  EREG_AAIND0, EREG_AAIND1, EREG_AAIND2, EREG_AAIND3, EREG_AAIND4, EREG_AAIND5, EREG_AAIND6, EREG_AAIND7, EREG_AAIND8, EREG_AAIND9,
  EREG_AAIND10, EREG_AAIND11, EREG_AAIND12, EREG_AAIND13, EREG_AAIND14, EREG_AAIND15,

  EREG_AAINCR0, EREG_AAINCR1, EREG_AAINCR2, EREG_AAINCR3, EREG_AAINCR4, EREG_AAINCR5, EREG_AAINCR6, EREG_AAINCR7,

  EREG_AALDI0,  EREG_AALDI1,  EREG_AALDI2,  EREG_AALDI3,  EREG_AALDI4,  EREG_AALDI5,  EREG_AALDI6,  EREG_AALDI7,  EREG_AALDI8,  EREG_AALDI9,
  EREG_AALDI10, EREG_AALDI11, EREG_AALDI12, EREG_AALDI13, EREG_AALDI14, EREG_AALDI15, EREG_AALDI16, EREG_AALDI17, EREG_AALDI18, EREG_AALDI19,
  EREG_AALDI20, EREG_AALDI21, EREG_AALDI22, EREG_AALDI23, EREG_AALDI24, EREG_AALDI25, EREG_AALDI26, EREG_AALDI27, EREG_AALDI28, EREG_AALDI29,
  EREG_AALDI30, EREG_AALDI31, EREG_AALDI32, EREG_AALDI33, EREG_AALDI34, EREG_AALDI35, EREG_AALDI36, EREG_AALDI37, EREG_AALDI38, EREG_AALDI39,
  EREG_AALDI40, EREG_AALDI41, EREG_AALDI42, EREG_AALDI43, EREG_AALDI44, EREG_AALDI45, EREG_AALDI46, EREG_AALDI47, EREG_AALDI48, EREG_AALDI49,
  EREG_AALDI50, EREG_AALDI51, EREG_AALDI52, EREG_AALDI53, EREG_AALDI54, EREG_AALDI55, EREG_AALDI56, EREG_AALDI57, EREG_AALDI58, EREG_AALDI59,
  EREG_AALDI60, EREG_AALDI61, EREG_AALDI62, EREG_AALDI63,

  EREG_AALDA0,  EREG_AALDA1,  EREG_AALDA2,  EREG_AALDA3,  EREG_AALDA4,  EREG_AALDA5,  EREG_AALDA6,  EREG_AALDA7,  EREG_AALDA8,  EREG_AALDA9,
  EREG_AALDA10, EREG_AALDA11, EREG_AALDA12, EREG_AALDA13, EREG_AALDA14, EREG_AALDA15, EREG_AALDA16, EREG_AALDA17, EREG_AALDA18, EREG_AALDA19,
  EREG_AALDA20, EREG_AALDA21, EREG_AALDA22, EREG_AALDA23, EREG_AALDA24, EREG_AALDA25, EREG_AALDA26, EREG_AALDA27, EREG_AALDA28, EREG_AALDA29,
  EREG_AALDA30, EREG_AALDA31, EREG_AALDA32, EREG_AALDA33, EREG_AALDA34, EREG_AALDA35, EREG_AALDA36, EREG_AALDA37, EREG_AALDA38, EREG_AALDA39,
  EREG_AALDA40, EREG_AALDA41, EREG_AALDA42, EREG_AALDA43, EREG_AALDA44, EREG_AALDA45, EREG_AALDA46, EREG_AALDA47, EREG_AALDA48, EREG_AALDA49,
  EREG_AALDA50, EREG_AALDA51, EREG_AALDA52, EREG_AALDA53, EREG_AALDA54, EREG_AALDA55, EREG_AALDA56, EREG_AALDA57, EREG_AALDA58, EREG_AALDA59,
  EREG_AALDA60, EREG_AALDA61, EREG_AALDA62, EREG_AALDA63,

  EREG_AAFSTR, EREG_AALDV, EREG_AALDM, EREG_AAUBUSYFLAGS,

  EREG_LSR, EREG_EMPTY_LO, EREG_EMPTY_HI,

  EREG_RET,

  EREG_LAST
};

//----------------------------------------------------------------------------//

enum WaitOpc {
  WAIT_ALL_C    = BIT(0),
  WAIT_ALL_E    = BIT(1), // NOT SUPPORTED
  WAIT_ST_C     = BIT(2),
  WAIT_LD_C     = BIT(3),
  WAIT_FL_C     = BIT(4), // NOT SUPPORTED
  WAIT_MA_C     = BIT(5), // NOT SUPPORTED
  WAIT_TRAP     = BIT(6), // NOT SUPPORTED
  WAIT_SAS      = BIT(7), // SUPPORTED in v5
  WAIT_SAL      = BIT(8), // SUPPORTED in v5
  WAIT_LAS      = BIT(9), // SUPPORTED in v6
  WAIT_LAL      = BIT(10),// SUPPORTED in v6
  WAIT_MT       = BIT(11),// SUPPORTED in v6

  //  fullFence: Ensures lack of reordering of loads or stores before the fence
  //  with loads or stores after the fence.
  WAIT_FENCE_C = (WAIT_LD_C | WAIT_ST_C)
};

enum {
  LD_MAS_NORMAL = 0x0,
  LD_MAS_SEMI_SPEC = 0x0,
  LD_MAS_DIS_TRANSL = 0x1,
  LD_MAS_SPEC_2ND = 0x1,
  LD_MAS_CHECK = 0x2,
  LD_MAS_CHECK_UNLOCK = 0x3,
  LD_MAS_SPEC = 0x3,
  LD_MAS_FILL = 0x4,
  LD_MAS_LOCK_CHECK = 0x4,
  LD_MAS_TRAP_S = 0x5,
  LD_MAS_TRAP_LS = 0x6,
  LD_MAS_LOCK_WAIT = 0x7,
  LD_MAS_SPEC_LOCK_CHECK = 0x7,
};

enum {
  ST_MAS_NORMAL = 0x0,
  ST_MAS_SEMI_SPEC = 0x0,
  ST_MAS_DIS_TRANSL = 0x1,
  ST_MAS_SPEC_2ND = 0x1,
  ST_MAS_WAIT_UNLOCK = 0x2,
  ST_MAS_STORE_UNLOCK = 0x3,
  ST_MAS_NO_CHECK = 0x4,
  ST_MAS_UNLOCK = 0x5,
  ST_MAS_IO_OP = 0x6,
  ST_MAS_SPECIAL = 0x7
};

enum {
  MAS_SPEC_CACHE_FLUSH = 0x0,
  MAS_SPEC_LINE_FLUSH = 0x1,
  MAS_SPEC_INST_LINE_FLUSH = 0x2,
  MAS_SPEC_TLB_PAGE_FLUSH = 0x2,
  MAS_SPEC_LOCKS_INVALIDATE = 0x3,
  MAS_SPEC_INST_FLUSH = 0x4,
  MAS_SPEC_TLB_FLUSH = 0x4,
  MAS_SPEC_TLB_ADDR_PROBE = 0x5,
  MAS_SPEC_TLB_ENTRY_PROBE = 0x6,
  MAS_SPEC_AAU_REG = 0x7,
  MAS_SPEC_MMU_REG = 0x8,
  MAS_SPEC_DTLB = 0x9,
  MAS_SPEC_L1CACHE = 0xa,
  MAS_SPEC_L2CACHE = 0xb,
  MAS_SPEC_SPECREG = 0xd,
};

enum {
  CT_TYPE_NORMAL = 0x0,
  CT_TYPE_LOOP_END = 0x4,
  CT_TYPE_LOOP_END_OR_PRED = 0x6,
  CT_TYPE_LOCK_COND = 0x8,
  CT_TYPE_PL_COND = 0x9,
  CT_TYPE_DT_COND = 0xa, // real value is 0x8
  CT_TYPE_NOT_LOOP_END_AND_PRED = 0xf,
  CT_TYPE_MASK = 0xF
};

enum {
  CT_PSRC_TYPE_CMP  = 0 << 4,
  CT_PSRC_TYPE_CMPj = 1 << 4,
  CT_PSRC_TYPE_CLP  = 2 << 4,
  CT_PSRC_TYPE_MASK = 3 << 4
};

enum {
  CT_CLP_NUM0 = 0x0,
  CT_CLP_NUM1 = 0x1,
  CT_CLP_NUM2 = 0x2,
  CT_CLP_MASK = 0x3
};

//----------------------------------------------------------------------------//
// pls    aa    nop ss   css als
// 16-18  15-12 11  10-8 7-6 5-0

#define EU_AL0 0x00001
#define EU_AL1 0x00002
#define EU_AL2 0x00004
#define EU_AL3 0x00008
#define EU_AL4 0x00010
#define EU_AL5 0x00020
#define EU_CS0 0x00040
#define EU_CS1 0x00080
#define EU_SS0 0x00100
#define EU_SS1 0x00200
#define EU_SS2 0x00400
#define EU_NOP 0x00800
#define EU_AA0 0x01000
#define EU_AA1 0x02000
#define EU_AA2 0x04000
#define EU_AA3 0x08000
#define EU_PL0 0x10000
#define EU_PL1 0x20000
#define EU_PL2 0x40000

enum ExecUnitMask {
  EU_AL_012345 = EU_AL0 | EU_AL1 | EU_AL2 | EU_AL3 | EU_AL4 | EU_AL5,
  EU_AL_03     = EU_AL0 | EU_AL3,
  EU_AL_14     = EU_AL1 | EU_AL4,
  EU_AL_0134   = EU_AL0 | EU_AL1 | EU_AL3 | EU_AL4,
  EU_AL_0235   = EU_AL0 | EU_AL2 | EU_AL3 | EU_AL5,
  EU_AL_25     = EU_AL2 | EU_AL5,
  EU_AL_5      = EU_AL5,
  EU_AL_2      = EU_AL2,
  EU_AL_1      = EU_AL1,
  EU_AL_0      = EU_AL0,
  EU_CS_0      = EU_CS0,
  EU_CS_0_AL_0134 = EU_CS0 | EU_AL_0134,
  EU_CS_1      = EU_CS1,
  EU_SS_CTOP   = EU_SS0,
  EU_SS_BAP    = EU_SS1,
  EU_SS_ABN    = EU_SS2,
  EU_HS_NOP    = EU_NOP,
  EU_PL_01     = EU_PL0 | EU_PL1,
  EU_PL_012    = EU_PL0 | EU_PL1 | EU_PL2,
  EU_AA_0      = EU_AA0,
  EU_AA_1      = EU_AA1,
  EU_AA_2      = EU_AA2,
  EU_AA_3      = EU_AA3,
  EU_AA_0123   = EU_AA0 | EU_AA1 | EU_AA2 | EU_AA3,
  EU_UNLOCK_ST = EU_AL2 | EU_AL5 | EU_CS0,
  EU_ALL       = EU_AL_012345 | EU_CS0 | EU_CS1 | EU_SS0 | EU_SS1 | EU_SS2 | EU_NOP | EU_PL0 | EU_PL1 | EU_PL2 | EU_AA_0123,
  EU_CLUSTER1  = EU_AL0 | EU_AL1 | EU_AL2,
  EU_CLUSTER2  = EU_AL3 | EU_AL4 | EU_AL5,
  EU_LAST      = 0xFFFFF
};

//----------------------------------------------------------------------------//

enum OutOpType {
  OOT_0I,
  OOT_1I,
  OOT_1A,
  OOT_2I,
  OOT_4I,
  OOT_1M,
  OOT_2M,
  OOT_4F,
  OOT_5F,
  OOT_6F,
  OOT_8R,
  OOT_8F,
  OOT_9F,
  OOT_10F,
  OOT_10_14F,
  OOT_11F,
  OOT_11_14F,
  OOT_11_16F,
  OOT_11_18F,
  OOT_12F,
  OOT_13F,
  OOT_14F,
  OOT_14_16F,
  OOT_16F,
  OOT_18F,
  OOT_31F,
  OOT_19_R,
  OOT_CT,  // ct insts.
  OOT_PRI, // predicate int.
  OOT_PRF, // predicate float.
  OOT_LAST
};

//----------------------------------------------------------------------------//

enum InOpType {
  IOT_I,
  IOT_F,
  IOT_S,
  IOT_W,    // IOT_S without bypass.
  IOT_R,
  IOT_LAST, // last register file type.
  IOT_CT,   // ct insts.
  IOT_PROC_CT
};

//----------------------------------------------------------------------------//

struct InstInfo {
  const char*  name;
  int32_t      src1;
  int32_t      src2;
  ExecUnitMask eu_mask;
  OutOpType    out_op_type;
  InOpType     in_op_type;
};

//----------------------------------------------------------------------------//

class Globals {
public:
  static InstCode     getLdOpcode(DataType);
  static InstCode     getStOpcode(DataType);
  static InstCode     getLdOpcode(int32_t);
  static InstCode     getStOpcode(int32_t);
  static InstCode     getFMAOpcode(InstCode op1, InstCode op2, bool reverse);
  static bool         isMemRead(InstCode);
  static bool         isMemWrite(InstCode);
  static bool         isSigned(DataType);             // is data signed
  static bool         isFloat(DataType);              // is data signed
  static int32_t      getDataSize(DataType);          // size in bytes
  static int32_t      getMemSize(InstCode opcode);    // size of mem operation
  static int8_t       getLatency(InstCode, InstCode, bool, int); // read after write latency (first arg is write)
  static int8_t       getMaxLatency(InstCode);        // Max latency after write inst.
  static ExecUnitMask getEuMask(InstCode);            //
  static int32_t      align(int32_t, int32_t);
  static int32_t      getSrc1(InstCode inst_code) { return inst_info[inst_code].src1; }
  static int32_t      getSrc2(InstCode inst_code) { return inst_info[inst_code].src2; }
  static int32_t      getLPredForAls(int32_t als);
public:
  static const InstInfo     inst_info[];                // instructions description
#ifdef TANGO_ASSERT
  static       uint64_t     inst_counters[];
#endif // TANGO_ASSERT

};

//----------------------------------------------------------------------------//


class TangoStream {
public:
  TangoStream();

  void resetOutputType(bool is_singlethread) { _single_thread = is_singlethread; }

	TangoStream& operator <<(bool);
	TangoStream& operator <<(float);
	TangoStream& operator <<(double);
	TangoStream& operator <<(int32_t);
	TangoStream& operator <<(int64_t);
	TangoStream& operator <<(uint32_t);
	TangoStream& operator <<(uint64_t);
	TangoStream& operator <<(const char*);
	TangoStream& operator <<(const std::string&);
	TangoStream& operator <<(const arena_string&);
	TangoStream& operator <<(const persistent_string&);

  bool _single_thread;
};

//----------------------------------------------------------------------------//

} // namespace tango

#endif //SHARE_VM_TANGO_TANGO_INST_HPP
